/*
 * OLSR Basic Multicast Forwarding (BMF) plugin.
 * Copyright (c) 2005, 2006, Thales Communications, Huizen, The Netherlands.
 * Written by Erik Tromp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * * Redistributions of source code must retain the above copyright 
 *   notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright 
 *   notice, this list of conditions and the following disclaimer in 
 *   the documentation and/or other materials provided with the 
 *   distribution.
 * * Neither the name of Thales, BMF nor the names of its 
 *   contributors may be used to endorse or promote products derived 
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* -------------------------------------------------------------------------
 * File       : Bmf.c
 * Description: Multicast forwarding functions
 * Created    : 29 Jun 2006
 *
 * $Id$ 
 *
 * $Log$
 * ------------------------------------------------------------------------- */

#define _MULTI_THREADED

#include "Bmf.h"

/* System includes */
#include <stdio.h> /* NULL */
#include <sys/types.h> /* ssize_t */
#include <string.h> /* strerror() */
#include <errno.h> /* errno */
#include <assert.h> /* assert() */
#include <linux/if_packet.h> /* struct sockaddr_ll, PACKET_MULTICAST */
#include <pthread.h> /* pthread_t, pthread_create() */
#include <signal.h> /* sigset_t, sigfillset(), sigdelset(), SIGINT */
#include <netinet/ip.h> /* struct ip */

/* OLSRD includes */
#include "defs.h" /* olsr_cnf */
#include "olsr.h" /* olsr_printf */
#include "scheduler.h" /* olsr_register_scheduler_event */
#include "mid_set.h" /* mid_lookup_main_addr() */
#include "mpr_selector_set.h" /* olsr_lookup_mprs_set() */
#include "link_set.h" /* get_best_link_to_neighbor() */

/* Plugin includes */
#include "NetworkInterfaces.h" /* TBmfInterface, CreateBmfNetworkInterfaces(), CloseBmfNetworkInterfaces() */
#include "Address.h" /* IsMulticast(), IsLocalBroadcast() */
#include "Packet.h" /* ETH_TYPE_OFFSET, IFHWADDRLEN etc. */
#include "PacketHistory.h" /* InitPacketHistory() */
#include "DropList.h" /* DropMac() */

static pthread_t BmfThread;
static int BmfThreadRunning = 0;


/* -------------------------------------------------------------------------
 * Function   : BmfPacketCaptured
 * Description: Handle a captured raw IP packet
 * Input      : intf - the network interface on which the packet was captured
 *              buffer - space for the encapsulation header, followed by
 *                the captured packet
 *              len - the number of octets in the encapsulation header plus
 *                captured packet
 * Output     : none
 * Return     : none
 * Data Used  : BmfInterfaces
 * Notes      : The packet is assumed to be captured on a socket of family
 *              PF_PACKET and type SOCK_RAW.
 * ------------------------------------------------------------------------- */
static void BmfPacketCaptured(struct TBmfInterface* intf, unsigned char* buffer, ssize_t len)
{
  unsigned char* srcMac;
  union olsr_ip_addr srcIp;
  union olsr_ip_addr destIp;
  union olsr_ip_addr* origIp;
  struct sockaddr_in encapDest;
  struct TBmfInterface* nextFwIntf;
  int isFromOlsrIntf; /* Boolean indicating if packet captured on OLSR-enabled interface */
  int iAmNotMpr;
  unsigned char* ethPkt = buffer + ENCAP_HDR_LEN;
  ssize_t ethPktLen = len - ENCAP_HDR_LEN;
  struct ip* ipData;

  /* Only forward IPv4 packets */
  u_int16_t type;
  memcpy(&type, ethPkt + ETH_TYPE_OFFSET, 2);
  if (ntohs(type) != IPV4_TYPE)
  {
    return;
  }

  ipData = (struct ip*)(ethPkt + IP_HDR_OFFSET);

  /* Only forward multicast or local broadcast packets */
  COPY_IP(&destIp, &ipData->ip_dst);
  if (! IsMulticast(&destIp)
#ifdef SVEN_OLA_DISABLED
     /* 
      * Sven-Ola@gmx.de: In a bigger mesh, there are a lot of
      * accidential bcast sources.  Disabled to save bandwidth 
      */
     && ! IsLocalBroadcast(&destIp, intf)
#endif
  )
  {
    return;
  }

  /* Discard OLSR packets (UDP port 698) and BMF encapsulated packets */
  if (IsOlsrOrBmfPacket(intf, ethPkt, ethPktLen))
  {
    return;
  }

  COPY_IP(&srcIp, &ipData->ip_src);
  olsr_printf(
    9,
    "%s: pkt of %d bytes ttl=%d incoming on \"%s\": %s->%s\n",
    PLUGIN_NAME_SHORT,
    ethPktLen,
    GetIpTtl(ethPkt),
    intf->ifName,
    olsr_ip_to_string(&srcIp),
    olsr_ip_to_string(&destIp));

  /* Apply drop list for testing purposes. */
  srcMac = ethPkt + IFHWADDRLEN;
  if (IsInDropList(srcMac))
  {
    olsr_printf(
      9,
      "%s: --> discarding: source MAC (%.2x:%.2x:%.2x:%.2x:%.2x:%.2x) found in drop list\n",
      PLUGIN_NAME_SHORT,
      srcMac, srcMac + 1, srcMac + 2, srcMac + 3, srcMac + 4, srcMac + 5);
    return;
  }

  /* Lookup main address of source in the MID table of OLSR */
  origIp = mid_lookup_main_addr(&srcIp);
  if (origIp == NULL)
  {
    origIp = &srcIp;
  }

  /* Check if this packet is received on an OLSR-enabled interface */
  isFromOlsrIntf = (intf->olsrIntf != NULL);

  /* If this packet is captured on a non-OLSR interface, it will be forwarded
   * only to OLSR interfaces, thereby crossing the boundary between the external
   * network and the OLSR network. So decrease the TTL and re-calculate the IP header
   * checksum. */
  if (! isFromOlsrIntf)
  {
    DecreaseTtlAndUpdateHeaderChecksum(ethPkt);
  }

  /* If the resulting TTL is <= 0, this packet life has ended, so do not forward it */
  if (GetIpTtl(ethPkt) <= 0)
  {
    olsr_printf(
      9,
      "%s: --> discarding: TTL=0\n",
      PLUGIN_NAME_SHORT);
    return;
  }

  /* Check if this packet was seen recently */
  u_int32_t crc32 = PacketCrc32(ethPkt, ethPktLen);
  if (CheckAndMarkRecentPacket(Hash16(crc32)))
  {
    olsr_printf(
      9,
      "%s: --> discarding: packet is duplicate\n",
      PLUGIN_NAME_SHORT);
    return;
  }

  /* Compose encapsulation header */
  struct TEncapHeader* encapHdr = (struct TEncapHeader*) buffer;
  memset (encapHdr, 0, ENCAP_HDR_LEN);
  encapHdr->crc32 = htonl(crc32);
  encapHdr->encapttl = GetIpTtl(ethPkt);

  /* If this packet is captured on an OLSR interface from an OLSR neighbor,
   * check with OLSR if I am MPR for that neighbor */
  iAmNotMpr =
    (isFromOlsrIntf /* captured on an OLSR interface */
    && get_best_link_to_neighbor(origIp) != NULL /* from an OLSR neighbor */
    && olsr_lookup_mprs_set(origIp) == NULL); /* but not selected as MPR */

  memset(&encapDest, 0, sizeof(encapDest));
  encapDest.sin_family = AF_INET;
  encapDest.sin_port = htons(BMF_ENCAP_PORT);

  /* Check with each interface what needs to be done on it */
  nextFwIntf = BmfInterfaces;
  while (nextFwIntf != NULL)
  {
    int isToOlsrIntf; /* Boolean indicating if forwarding interface is OLSR-enabled */

    struct TBmfInterface* fwIntf = nextFwIntf;
    nextFwIntf = fwIntf->next;

    isToOlsrIntf = (fwIntf->olsrIntf != NULL);

    /* The following if-else statement seems very complicated. Let's
     * draw a Karnaugh-diagram of it for some clarification.
     *
     *                -------  isFromOlsrIntf 
     *           --------      isToOlsrIntf
     *       +---+---+---+---+
     *       | 3 | 2 | 2 | 1 |
     *       +---+---+---+---+
     *     | | X | X | 4 | 1 |
     *       +---+---+---+---+
     * iAmNotMpr
     *
     * X = does not occur (iAmNotMpr = isFromOlsrIntf && ...)
     * 1 = handled in first if-clause: if (isFromOlsrIntf && !isToOlsrIntf)...
     * 2 = handled in second if-clause: if (isToOlsrIntf && !iAmNotMpr)...
     * 3 = handled in third if-clause: if (!isFromOlsrIntf && !isToOlsrIntf)...
     * 4 = handled in else-clause.
     */

    /* Forward from OLSR-interface to non-OLSR interface */
    if (isFromOlsrIntf && !isToOlsrIntf)
    {
      struct TSaveTtl sttl;

      /* Save IP header checksum and the TTL-value of the packet */ 
      SaveTtlAndChecksum(ethPkt, &sttl);

      /* Save IP header checksum and the TTL-value of the packet, then
       * decrease the TTL by 1 before writing */
      DecreaseTtlAndUpdateHeaderChecksum(ethPkt);

      /* If the TTL is <= 0, do not forward this packet */
      if (GetIpTtl(ethPkt) <= 0)
      {
        olsr_printf(
          9,
          "%s: --> not forwarding on \"%s\": TTL=0\n",
          PLUGIN_NAME_SHORT,
          fwIntf->ifName);
      }
      else
      {
        /* Change source MAC address to that of sending interface */
        memcpy(ethPkt + IFHWADDRLEN, fwIntf->macAddr, IFHWADDRLEN);

        int nBytesWritten = write(fwIntf->capturingSkfd, ethPkt, ethPktLen);
        if (nBytesWritten != ethPktLen)
        {
          olsr_printf(
            1,
            "%s: write() error forwarding pkt for %s to \"%s\": %s\n",
            PLUGIN_NAME,
            olsr_ip_to_string(&destIp),
            fwIntf->ifName,
            strerror(errno));
        }
        else
        {
          olsr_printf(
            9,
            "%s: --> forwarded(capt) ttl=%d to \"%s\"\n",
            PLUGIN_NAME_SHORT,
            GetIpTtl(ethPkt),
            fwIntf->ifName);
        }
      }

      /* Restore the IP header checksum and the TTL-value of the packet */
      RestoreTtlAndChecksum(ethPkt, &sttl);

    } /* if (isFromOlsrIntf && !isToOlsrIntf) */

    /* Forward from any interface to OLSR interface. Packets from non-OLSR interfaces
     * already had their TTL decreased. */
    else /* !isFromOlsrIntf || isToOlsrIntf */
    if (isToOlsrIntf && !iAmNotMpr)
    {
      int nBytesWritten;
      unsigned char ttl;

      /* Change encapsulated source MAC address to that of sending interface */
      memcpy(ethPkt + IFHWADDRLEN, fwIntf->macAddr, IFHWADDRLEN);

      /* Destination address is local broadcast */
      encapDest.sin_addr.s_addr = ((struct sockaddr_in*)&fwIntf->olsrIntf->int_broadaddr)->sin_addr.s_addr;

      /* SVEN_OLA: Normal TTL of this socket will be 64 if unset */
      ttl = GetIpTtl(ethPkt);
      if (0 <= (nBytesWritten = setsockopt(fwIntf->encapsulatingSkfd, IPPROTO_IP, IP_TTL, &ttl, sizeof(ttl))))

      nBytesWritten = sendto(
        fwIntf->encapsulatingSkfd,
        buffer,
        len,
        MSG_DONTROUTE,
        (struct sockaddr*) &encapDest,
        sizeof(encapDest));                   

      /*
       * Sven-Ola@gmx.de: Very primitive testing of forward error correction
       */
      if (nBytesWritten == len)
        nBytesWritten = sendto(
          fwIntf->encapsulatingSkfd,
          buffer,
          len,
          MSG_DONTROUTE,
          (struct sockaddr*) &encapDest,
          sizeof(encapDest));

      if (nBytesWritten != len)
      {
        olsr_printf(
          1,
          "%s: sendto() error forwarding pkt for %s to \"%s\": %s\n",
          PLUGIN_NAME,
          olsr_ip_to_string(&destIp),
          fwIntf->ifName,
          strerror(errno));
      }
      else
      {
        olsr_printf(
          9,
          "%s: --> encapsulated and forwarded ttl=%d to \"%s\"\n",
          PLUGIN_NAME_SHORT,
          GetIpTtl(ethPkt),
          fwIntf->ifName);
      } /* if (nBytesWritten != len) */
    } /* else if (isToOlsrIntf && !iAmNotMpr) */

    else /* (!isFromOlsrIntf || isToOlsrIntf) && (!isToOlsrIntf || iAmNotMpr) */
    if (!isFromOlsrIntf && !isToOlsrIntf)
    {
      olsr_printf(
        9,
        "%s: --> not forwarding from \"%s\" to \"%s\": both non-OLSR interfaces\n",
        PLUGIN_NAME_SHORT,
        intf->ifName,
        fwIntf->ifName);
    }

    else /* (!isFromOlsrIntf || isToOlsrIntf) && (!isToOlsrIntf || iAmNotMpr) && (isFromOlsrIntf || isToOlsrIntf) */
    {
      olsr_printf(
        9,
        "%s: --> not forwarding from \"%s\" to \"%s\": I am not selected as MPR by %s\n",
        PLUGIN_NAME_SHORT,
        intf->ifName,
        fwIntf->ifName,
        olsr_ip_to_string(&srcIp));
    }
  } /* while (nextFwIntf != NULL) */
}

/* -------------------------------------------------------------------------
 * Function   : BmfEncapsulatedPacketReceived
 * Description: Handle a received BMF-encapsulated IP packet
 * Input      : intf - the network interface on which the packet was received
 *              fromIp - the IP node that forwarded the packet to us
 *              buffer - the received encapsulated packet
 *              len - the number of octets in the received encapsulated packet
 * Output     : none
 * Return     : none
 * Data Used  : BmfInterfaces
 * Notes      : The packet is assumed to be received on a socket of family
 *              PF_INET and type SOCK_DGRAM (UDP).
 * ------------------------------------------------------------------------- */
static void BmfEncapsulatedPacketReceived(
  struct TBmfInterface* intf, 
  union olsr_ip_addr* fromIp,
  unsigned char* buffer,
  ssize_t len)
{
  union olsr_ip_addr* forwarder;
  int nBytesToWrite;
  unsigned char* bufferToWrite;
  int nBytesWritten;
  int iAmMpr;
  struct sockaddr_in encapDest;
  struct TBmfInterface* nextFwIntf;
  struct ip* ipData;
  unsigned char* ethPkt;
  ssize_t ethPktLen;

  /* Are we talking to ourselves? */
  if (if_ifwithaddr(fromIp) != NULL)
  {
    return;
  }

  /* Encapsulated packet received on non-OLSR interface? Then discard */
  if (intf->olsrIntf == NULL)
  {
    return;
  }

  /* Apply drop list? No, not needed: encapsulated packets are routed,
   * so filtering should be done by adding a rule to the iptables FORWARD
   * chain, e.g.:
   * iptables -A FORWARD -m mac --mac-source 00:0C:29:28:0E:CC -j DROP */

  ethPkt = buffer + ENCAP_HDR_LEN;
  ethPktLen = len - ENCAP_HDR_LEN;

  ipData = (struct ip*) (ethPkt + IP_HDR_OFFSET);

  olsr_printf(
    9,
    "%s: encapsulated pkt of %d bytes ttl=%d incoming on \"%s\": %s->",
    PLUGIN_NAME_SHORT,
    ethPktLen,
    GetIpTtl(ethPkt),
    intf->ifName,
    inet_ntoa(ipData->ip_src));
  olsr_printf(
    9,
    "%s, ",
    inet_ntoa(ipData->ip_dst)); /* not possible to call inet_ntoa twice in same printf */
  olsr_printf(
    9,
    "forwarded by %s\n",
    olsr_ip_to_string(fromIp));

  /* Get encapsulation header */
  struct TEncapHeader* encapHdr = (struct TEncapHeader*) buffer;

  /* Check if this packet was seen recently */
  if (CheckAndMarkRecentPacket(Hash16(ntohl(encapHdr->crc32))))
  {
    olsr_printf(
      9,
      "%s: --> discarding: packet is duplicate\n",
      PLUGIN_NAME_SHORT);
    return;
  }

  /* Unpack encapsulated packet and send a copy to myself via the EtherTunTap device */
  bufferToWrite = ethPkt;
  nBytesToWrite = ethPktLen;
  if (TunOrTap == TT_TUN)
  {
    bufferToWrite += IP_HDR_OFFSET;
    nBytesToWrite -= IP_HDR_OFFSET;
  }
  nBytesWritten = write(EtherTunTapFd, bufferToWrite, nBytesToWrite);
  if (nBytesWritten != nBytesToWrite)
  {
    olsr_printf(
      1,
      "%s: write() error forwarding encapsulated pkt to \"%s\": %s\n",
      PLUGIN_NAME,
      EtherTunTapIfName,
      strerror(errno));
  }
  else
  {
    olsr_printf(
      9,
      "%s: --> unpacked and forwarded ttl=%d to \"%s\"\n",
      PLUGIN_NAME_SHORT,
      GetIpTtl(ethPkt),
      EtherTunTapIfName);
  }
  
  /*
   * Sven-Ola@gmx.de: The original implemenation make a wave
   * through the complete mesh for every packet. Packets are
   * discarded if any device has seen it (most bad case). We
   * Want some "local" distribution mode, so I've added some
   * hopcounter here - derived from the original ttl
   */
  if (0 == encapHdr->encapttl || 0 == --encapHdr->encapttl) {
    olsr_printf(
      9,
      "%s: --> dicarding encapsulated: TTL=0\n",
      PLUGIN_NAME_SHORT);
    return;
  }

  /* Lookup main address of forwarding node */
  forwarder = mid_lookup_main_addr(fromIp);
  if (forwarder == NULL)
  {
    forwarder = fromIp;
  }

  /* Check if I am MPR for the forwarder */
  iAmMpr = (olsr_lookup_mprs_set(forwarder) != NULL);

  memset(&encapDest, 0, sizeof(encapDest));
  encapDest.sin_family = AF_INET;
  encapDest.sin_port = htons(BMF_ENCAP_PORT);

  nextFwIntf = BmfInterfaces;
  while (nextFwIntf != NULL)
  {
    struct TBmfInterface* fwIntf = nextFwIntf;
    nextFwIntf = fwIntf->next;

    /* Forward from OLSR interface to non-OLSR interface: unpack encapsulated
     * packet, decrease TTL and forward */
    if (fwIntf->olsrIntf == NULL)
    {
      struct TSaveTtl sttl;

      /* Save IP header checksum and the TTL-value of the packet, then 
       * decrease the TTL by 1 before writing */
      SaveTtlAndChecksum(ethPkt, &sttl);
      DecreaseTtlAndUpdateHeaderChecksum(ethPkt);

      /* If the TTL is <= 0, do not forward this packet */
      if (GetIpTtl(ethPkt) <= 0)
      {
        olsr_printf(
          9,
          "%s: --> not forwarding on \"%s\": TTL=0\n",
          PLUGIN_NAME_SHORT,
          fwIntf->ifName);
      }
      else
      {
        nBytesWritten = write(fwIntf->capturingSkfd, ethPkt, ethPktLen);
        if (nBytesWritten != ethPktLen)
        {
          olsr_printf(
            1,
            "%s: write() error forwarding unpacked encapsulated MC pkt to \"%s\": %s\n",
            PLUGIN_NAME,
            fwIntf->ifName,
            strerror(errno));
        }
        else
        {
          olsr_printf(
            9,
            "%s: --> unpacked and forwarded ttl=%d, sttl=%d to \"%s\"\n",
            PLUGIN_NAME_SHORT,
            GetIpTtl(ethPkt),
            sttl.ttl,
            fwIntf->ifName);
        }
      }

      /* Restore the IP header checksum and the TTL-value of the packet */
      RestoreTtlAndChecksum(ethPkt, &sttl);

    } /* if (fwIntf->olsrIntf == NULL) */

    /* Forward from OLSR interface to OLSR interface: forward the packet if this
     * node is selected as MPR by the forwarding node */
    else if (iAmMpr)
    {
      /* Change source MAC address to that of sending interface */
      memcpy(buffer + IFHWADDRLEN, fwIntf->macAddr, IFHWADDRLEN);

      /* Destination address is local broadcast */
      encapDest.sin_addr.s_addr = ((struct sockaddr_in*)&fwIntf->olsrIntf->int_broadaddr)->sin_addr.s_addr;

      nBytesWritten = sendto(
        fwIntf->encapsulatingSkfd,
        buffer,
        len,
        MSG_DONTROUTE,
        (struct sockaddr*) &encapDest,
        sizeof(encapDest));                   

      if (nBytesWritten != len)
      {
        olsr_printf(
          1,
          "%s: sendto() error forwarding encapsulated pkt via \"%s\": %s\n",
          PLUGIN_NAME,
          fwIntf->ifName,
          strerror(errno));
      }
      else
      {
        olsr_printf(
          9,
          "%s: --> forwarded(encrec) ttl=%d to \"%s\"\n",
          PLUGIN_NAME_SHORT,
          GetIpTtl(ethPkt),
          fwIntf->ifName);
      }
    }  /* else if (iAmMpr) */
    else /* fwIntf->olsrIntf != NULL && !iAmMpr */
    {
      /* fwIntf is an OLSR interface and I am not selected as MPR */
      olsr_printf(
        9,
        "%s: --> not forwarding to \"%s\": I am not selected as MPR by %s\n",
        PLUGIN_NAME_SHORT,
        fwIntf->ifName,
        olsr_ip_to_string(fromIp));
    }
  } /* while (nextFwIntf != NULL) */
}

/* -------------------------------------------------------------------------
 * Function   : DoBmf
 * Description: Wait (blocking) for IP packets, then call the handler for each
 *              received packet
 * Input      : none
 * Output     : none
 * Return     : none
 * Data Used  : BmfInterfaces
 * ------------------------------------------------------------------------- */
static void DoBmf(void)
{
#define BUFFER_MAX 2048
  struct TBmfInterface* currIf;
  int nFdBitsSet;

  /* Compose set of socket file descriptors. 
   * Keep the highest descriptor seen. */
  int highestSkfd = -1;
  fd_set input_set;
  FD_ZERO(&input_set);

  currIf = BmfInterfaces;
  while (currIf != NULL)
  {
    FD_SET(currIf->capturingSkfd, &input_set);
    if (currIf->capturingSkfd > highestSkfd)
    {
      highestSkfd = currIf->capturingSkfd;
    }

    if (currIf->encapsulatingSkfd >= 0)
    {
      FD_SET(currIf->encapsulatingSkfd, &input_set);
      if (currIf->encapsulatingSkfd > highestSkfd)
      {
        highestSkfd = currIf->encapsulatingSkfd;
      }
    }

    currIf = currIf->next;    
  }

  assert(highestSkfd >= 0);

  /* Wait (blocking) for packets received on any of the sockets.
   * NOTE: don't use a timeout (last parameter). It causes a high system CPU load! */
  nFdBitsSet = select(highestSkfd + 1, &input_set, NULL, NULL, NULL);
  if (nFdBitsSet < 0)
  {
    if (errno != EINTR)
    {
      olsr_printf(1, "%s: select() error: %s\n", PLUGIN_NAME, strerror(errno));
    }
    return;
  }

  if (nFdBitsSet == 0)
  {
    /* No packets waiting. This is unexpected; normally we would excpect select(...)
     * to return only if at least one packet was received (so nFdBitsSet > 0), or
     * if this thread received a signal (so nFdBitsSet < 0). */
    return;
  }

  while (nFdBitsSet > 0)
  {
    struct TBmfInterface* nextIf = BmfInterfaces;
    while (nextIf != NULL)
    {
      int skfd;
      currIf = nextIf;
      nextIf = currIf->next;

      skfd = currIf->capturingSkfd;
      if (FD_ISSET(skfd, &input_set))
      {
        unsigned char buffer[BUFFER_MAX];
        struct sockaddr_ll pktAddr;
        socklen_t addrLen;
        int nBytes;
        unsigned char* ethPkt = buffer + ENCAP_HDR_LEN;
        struct ip* ipData;

        /* A packet was captured. */

        nFdBitsSet--;

        memset(&pktAddr, 0, sizeof(struct sockaddr_ll));
        addrLen = sizeof(pktAddr);

        /* Receive the packet, leaving space for the BMF encap header */
        nBytes = recvfrom(
          skfd,
          ethPkt,
          BUFFER_MAX - ENCAP_HDR_LEN,
          0,
          (struct sockaddr*)&pktAddr,
          &addrLen);
        if (nBytes < 0)
        {
          olsr_printf(
            1,
            "%s: recvfrom() error on \"%s\": %s\n",
            PLUGIN_NAME,
            currIf->ifName,
            strerror(errno));
        }
        else
        {
          /* Don't let BMF crash by sending too short packets */
          int ipHeaderLen;
          ipData = (struct ip*) (ethPkt + IP_HDR_OFFSET);
          ipHeaderLen = ipData->ip_hl << 2;
          if (nBytes >= IP_HDR_OFFSET + ipHeaderLen)
          {
            if (pktAddr.sll_pkttype == PACKET_OUTGOING)
            {
              union olsr_ip_addr destIp;
              COPY_IP(&destIp, &ipData->ip_dst);
              if (IsMulticast(&destIp)
#ifdef SVEN_OLA_DISABLED
                 /* 
                  * Sven-Ola@gmx.de: In a bigger mesh, there are a lot of
                  * accidential bcast sources.  Disabled to save bandwidth 
                  */
                 || IsLocalBroadcast(&destIp, currIf)
#endif
              )
              {
                if (! IsOlsrOrBmfPacket(currIf, ethPkt, nBytes))
                {
                  /* For outbound packets, just record the fact that the packet was
                   * seen recently */
                  u_int32_t crc32 = PacketCrc32(ethPkt, nBytes);
                  MarkRecentPacket(Hash16(crc32));
                }
              }
            }
            else if (pktAddr.sll_pkttype == PACKET_MULTICAST
#ifdef SVEN_OLA_DISABLED
                 /* 
                  * Sven-Ola@gmx.de: In a bigger mesh, there are a lot of
                  * accidential bcast sources.  Disabled to save bandwidth 
                  */
                 || pktAddr.sll_pkttype == PACKET_BROADCAST
#endif
            )
            {
              /* An inbound multicast or broadcast packet was captured */
              BmfPacketCaptured(currIf, buffer, nBytes + ENCAP_HDR_LEN);
            }
          } /* if (nBytes >= IP_HDR_OFFSET + ipHeaderLen) */
        } /* if (nBytes < 0) */
      } /* if (FD_ISSET...) */

      skfd = currIf->encapsulatingSkfd;
      if (skfd >= 0 && (FD_ISSET(skfd, &input_set)))
      {
        unsigned char buffer[BUFFER_MAX];
        struct sockaddr_in from;
        socklen_t fromLen = sizeof(from);
        int nBytes;
        struct ip* ipData;

        /* An encapsulated packet was received */

        nFdBitsSet--;

        nBytes = recvfrom(skfd, buffer, BUFFER_MAX, 0, (struct sockaddr*)&from, &fromLen);
        if (nBytes < 0)
        {
          olsr_printf(
            1,
            "%s: recvfrom() error on \"%s\": %s\n",
            PLUGIN_NAME,
            currIf->ifName,
            strerror(errno));
        }
        else
        {
          /* Don't let BMF crash by sending too short packets. */
          int ipHeaderLen;
          ipData = (struct ip *) (buffer + IP_HDR_OFFSET);
          ipHeaderLen = ipData->ip_hl << 2;
          if (nBytes >= IP_HDR_OFFSET + ipHeaderLen)
          {
            union olsr_ip_addr srcIp;
            COPY_IP(&srcIp, &from.sin_addr.s_addr);
            BmfEncapsulatedPacketReceived(currIf, &srcIp, buffer, nBytes);
          }
          else
          {
            olsr_printf(
              1,
              "%s: encapsulated packet too short (%d bytes) from %s on \"%s\"\n",
              PLUGIN_NAME,
              nBytes,
              inet_ntoa(from.sin_addr),
              currIf->ifName);
          }
        } /* if (nBytes < 0) */
      } /* if (skfd >= 0 && (FD_ISSET...) */
    } /* while (intf != NULL) */
  } /* while (nFdBitsSet > 0) */
}

/* -------------------------------------------------------------------------
 * Function   : BmfSignalHandler
 * Description: Dummy signal handler function
 * Input      : signo - signal being handled
 * Output     : none
 * Return     : none
 * Data Used  : none
 * ------------------------------------------------------------------------- */
static void BmfSignalHandler(int signo)
{
  /* Dummy handler function */
  return;
}

/* -------------------------------------------------------------------------
 * Function   : BmfRun
 * Description: Receiver thread function
 * Input      : useless - not used
 * Output     : none
 * Return     : not used
 * Data Used  : BmfThreadRunning
 * Notes      : Another thread can gracefully stop this thread by writing a
 *              '0' into global variable 'BmfThreadRunning' followed by sending
 *              a SIGALRM signal.
 * ------------------------------------------------------------------------- */
static void* BmfRun(void* useless)
{
  /* Mask all signals except SIGALRM */
  sigset_t blockedSigs;
  sigfillset(&blockedSigs);
  sigdelset(&blockedSigs, SIGALRM);
  if (pthread_sigmask(SIG_BLOCK, &blockedSigs, NULL) < 0)
  {
    olsr_printf(1, "%s: pthread_sigmask() error: %s\n", PLUGIN_NAME, strerror(errno));
  }

  /* Set up the signal handler for the process: use SIGALRM to terminate
   * the BMF thread. Only if a signal handler is specified, does a blocking
   * system call return with errno set to EINTR; if a signal hander is not
   * specified, any system call in which the thread may be waiting will not
   * return. Note that the BMF thread is usually blocked in the select()
   * function (see DoBmf()). */
  if (signal(SIGALRM, BmfSignalHandler) == SIG_ERR)
  {
    olsr_printf(1, "%s: signal() error: %s\n", PLUGIN_NAME, strerror(errno));
  }

  /* Call the thread function until flagged to exit */
  while (BmfThreadRunning != 0)
  {
    DoBmf();
  }

  return NULL;
}

/* -------------------------------------------------------------------------
 * Function   : InterfaceChange
 * Description: Callback function passed to OLSRD for it to call whenever a
 *              network interface has been added, removed or updated
 * Input      : interf - the network interface to deal with
 *              action - indicates if the specified network interface was
 *              added, removed or updated.
 * Output     : none
 * Return     : always 0
 * Data Used  : none
 * ------------------------------------------------------------------------- */
int InterfaceChange(struct interface* interf, int action)
{
  switch (action)
  {
  case (IFCHG_IF_ADD):
    AddInterface(interf);
    olsr_printf(1, "%s: interface %s added\n", PLUGIN_NAME, interf->int_name);
    break;

  case (IFCHG_IF_REMOVE):
    CloseBmf();
    InitBmf(interf);
    olsr_printf(1, "%s: interface %s removed\n", PLUGIN_NAME, interf->int_name);
    break;

  case (IFCHG_IF_UPDATE):
    olsr_printf(1, "%s: interface %s updated\n", PLUGIN_NAME, interf->int_name);
    break;
      
  default:
    break;
  }
  return 0;
}

/* -------------------------------------------------------------------------
 * Function   : InitBmf
 * Description: Initialize the BMF plugin
 * Input      : skipThisIntf - specifies which network interface should not
 *              be enabled for BMF. Pass NULL to indicate all.
 * Output     : none
 * Return     : fail (0) or success (1)
 * Data Used  : BmfThreadRunning, BmfThread
 * ------------------------------------------------------------------------- */
int InitBmf(struct interface* skipThisIntf)
{
  if (CreateBmfNetworkInterfaces(skipThisIntf) < 0)
  {
    olsr_printf(1, "%s: Could not initialize any network interface!\n", PLUGIN_NAME);
    /* Continue anyway; maybe an interface will be added later */
  }

  /* Start running the multicast packet processing thread */
  BmfThreadRunning = 1;
  if (pthread_create(&BmfThread, NULL, BmfRun, NULL) == 0)
  {
    return 1;
  }
  return 0;
}

/* -------------------------------------------------------------------------
 * Function   : CloseBmf
 * Description: Close the BMF plugin and clean up
 * Input      : none
 * Output     : none
 * Return     : none
 * Data Used  : BmfThreadRunning, BmfThread
 * ------------------------------------------------------------------------- */
void CloseBmf()
{
  /* Signal BmfThread to exit */
  BmfThreadRunning = 0;
  if (pthread_kill(BmfThread, SIGALRM) < 0)
  /* Strangely enough, all running threads receive the SIGALRM signal. But only the
   * BMF thread is affected by this signal, having specified a handler for this
   * signal in its thread entry function BmfRun(...). */
  {
    olsr_printf(1, "%s: pthread_kill() error: %s\n", PLUGIN_NAME, strerror(errno));
  }

  /* Wait for BmfThread to acknowledge */
  if (pthread_join(BmfThread, NULL) < 0)
  {
    olsr_printf(1, "%s: pthread_join() error: %s\n", PLUGIN_NAME, strerror(errno));
  }

  /* Time to clean up */
  CloseBmfNetworkInterfaces();
}

/* -------------------------------------------------------------------------
 * Function   : RegisterBmfParameter
 * Description: Register a configuration parameter with the BMF process
 * Input      : key - the parameter name: "DropMac" or "NonOlsrIf"
 *              value - the parameter value
 * Output     : none
 * Return     : fatal error (<0), minor error (0) or success (>0)
 * Data Used  : none
 * ------------------------------------------------------------------------- */
int RegisterBmfParameter(char* key, char* value)
{
  if (strcmp(key, "DropMac") == 0)
  {
    return DropMac(value);
  }
  else if (strcmp(key, "NonOlsrIf") == 0)
  {
    return AddNonOlsrBmfIf(value);
  }

  /* Key not recognized */
  return 0;
}
